# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.reader.xlsx.sensitivity_analysis_reader import SensitivityAnalysisReader
from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer
from availsim4core.src.sensitivity_analysis.exploration_strategy_factory import ExplorationStrategyFactory, \
    SensitivityAnalysisStrategyError
from availsim4core.src.sensitivity_analysis.sensitivity_analysis import SensitivityAnalysis
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_zip_strategy import \
    ZipStrategyError


class test_ExplorationStrategyFactory(unittest.TestCase):

    def test_build__inner(self):
        parameter_name = "param_test"
        float_values = [1., 2.]
        strategy = "iNner"
        expected_resut = Inner(parameter_name, float_values)

        result = ExplorationStrategyFactory.build(parameter_name,
                                                  float_values,
                                                  strategy)

        self.assertEqual(expected_resut, result)

    def test_build__outer(self):
        parameter_name = "param_test"
        float_values = [1., 2.]
        strategy = "oUter"
        expected_resut = Outer(parameter_name, float_values)

        result = ExplorationStrategyFactory.build(parameter_name,
                                                  float_values,
                                                  strategy)

        self.assertEqual(expected_resut, result)

    def test_build__exception(self):
        parameter_name = "param_test"
        float_values = [1., 2.]
        strategy = "random_name_to_raise_exception"

        self.assertRaises(SensitivityAnalysisStrategyError, ExplorationStrategyFactory.build,
                          parameter_name,
                          float_values,
                          strategy)

    def test_build__zip_construction_exception(self):

        sensitivity_analysis_dictionary = {'SENSITIVITY_ANALYSIS':
            {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": f'[1, 2]',
                    "EXPLORATION_STRATEGY": "ZIP"},
                1: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": f'[1, 2, 3]',
                    "EXPLORATION_STRATEGY": "ZIP"}
            }
        }

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            sensitivity_analysis = SensitivityAnalysis(None,
                                                       None)

            self.assertRaises(ZipStrategyError, sensitivity_analysis.generate_analysis_list,
                                "fake_path")


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class ExplorationStrategy:
    """
    Represents the available strategies for a {sensitivity_analysis}.
    """

    def __init__(self, parameter_name, values):
        self.parameter_name = parameter_name
        self.values = values

    def __iter__(self):
        for value in self.values:
            yield value

    def __eq__(self, other):
        return self.parameter_name == other.parameter_name \
               and self.values == other.values

    def __repr__(self):
        return f"ExplorationStrategy {self.__class__.__name__}: {self.parameter_name}, values {self.values} "


class Inner(ExplorationStrategy):
    def __init__(self, parameter_name, values):
        super().__init__(parameter_name, values)


class Outer(ExplorationStrategy):
    def __init__(self, parameter_name, values):
        super().__init__(parameter_name, values)

class Zip(ExplorationStrategy):
    def __init__(self, parameter_name, values):
        super().__init__(parameter_name, values)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.discrete_event_simulation.event.c_event.component_c_event import ComponentCEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class ReevaluateOrderAllFailureEvent(ComponentCEvent):
    """
    class used to regenerate all FailureEvents at once, once they have all been executed
    C events are generated which themselves generate B events
    """
    __slots__ = 'event'

    def __init__(self,
                 priority: int,
                 context: Context,
                 component: Component,
                 event: Event):
        super().__init__(priority, context, component)
        self.event = event

    def __eq__(self, other):
        if not isinstance(self, ReevaluateOrderAllFailureEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event

    def __hash__(self):
        return hash((type(self), self.priority, self.component, self.event))

    def __str__(self):
        return f"ReevaluateOrderAllFailureEvent:: " \
               f"priority:{self.priority} - " \
               f"component:{self.component.name}"

    def generate_b_events(self, absolute_simulation_time):
        from availsim4core.src.discrete_event_simulation.b_event_generator import BEventGenerator
        return BEventGenerator.get_b_events_failure(absolute_simulation_time,
                                                    self.context,
                                                    self.component)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import re

import setuptools

VERSIONFILE = "availsim4core/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIREMENTS: dict = {
    'core': [
        'numpy',
        'pandas',
        'datetime',
        'matplotlib',
        'openpyxl',
        'scipy',
        'xlsxwriter',
        'qmcpy',
        'matplotlib',
        'networkx',
        'pygraphviz'
    ],
    'test': [
        'pytest',
        'memory_profiler',
        'Pillow',
        'pytest-cov',
    ],
    'dev': [
        # 'requirement-for-development-purposes-only',
    ],
    'doc': [
        'sphinx',
        'sphinx-glpi-theme',
        'sphinx-autoapi',
        'sphinxcontrib.napoleon',
        'sphinx-autodoc-typehints',
    ],
}

setuptools.setup(
    name="availsim4",
    version=verstr,
    author="TE-MPE",
    author_email="te-dep-mpe-ms@cern.ch",
    license='gpl-3.0',
    description="Availsim4 is a tool to predict reliability and availability "
                "of modern particle accelerators and their related systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cern.ch/availsim4/availsim4",
    python_requires='>=3.7, <4',
    setup_requires=['wheel'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=REQUIREMENTS['core'],
    extras_require={
        **REQUIREMENTS,
        # The 'dev' extra is the union of 'test' and 'doc', with an option
        # to have explicit development dependencies listed.
        'dev': [req
                for extra in ['dev', 'test', 'doc']
                for req in REQUIREMENTS.get(extra, [])],
        # The 'all' extra is the union of all requirements.
        'all': [req for reqs in REQUIREMENTS.values() for req in reqs],
    },
    packages=setuptools.find_packages(),
    package_data={'availsim4core': ['logging/logging.conf', 'version.txt']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'availsim4=availsim4core.main:main',
        ],
    },
)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.timeline.record_component import RecordComponent
from availsim4core.src.timeline.record_generator import RecordGenerator
from availsim4core.src.timeline.record_phase import RecordPhase


class test_RecordGenerator(unittest.TestCase):

    def test_get_initial_records(self):
        phase_set = {PhaseManager.NO_PHASE}
        rca_set = set()

        basic_1 = Basic(1, "test_basic_1", 0, [], [], [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], [], [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)

        context = Context(compound, PhaseManager(phase_set, set()), rca_set)

        result = RecordGenerator.get_initial_records(0, context)

        expected_record_list = [RecordPhase(PhaseManager.NO_PHASE,
                                            0, RecordGenerator.DESCRIPTION_INIT),
                                RecordComponent(basic_1,
                                                Status.RUNNING,
                                                0,
                                                PhaseManager.NO_PHASE,
                                                RecordGenerator.DESCRIPTION_INIT),
                                RecordComponent(basic_2,
                                                Status.RUNNING,
                                                0,
                                                PhaseManager.NO_PHASE,
                                                RecordGenerator.DESCRIPTION_INIT),
                                RecordComponent(compound,
                                                Status.RUNNING,
                                                0,
                                                PhaseManager.NO_PHASE,
                                                RecordGenerator.DESCRIPTION_INIT)
                                ]
        self.assertEqual(result, expected_record_list)

    def test_get_initial_records__phases(self):
        phase_first = Phase("phase_test", DeterministicLaw(10), True)
        phase_second = Phase("phase_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)
        phase_set = {phase_first, phase_second}

        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])

        context = Context(compound, PhaseManager(phase_set, set()), set())

        result = RecordGenerator.get_initial_records(0, context)

        expected_record_list = [RecordPhase(phase_first,
                                            0, RecordGenerator.DESCRIPTION_INIT),
                                RecordComponent(compound,
                                                Status.RUNNING,
                                                0,
                                                phase_first,
                                                RecordGenerator.DESCRIPTION_INIT)
                                ]
        self.assertEqual(result, expected_record_list)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from typing import List

from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.children_logic.children_logic_factory import ChildrenLogicFactory
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.component import Component, ComponentType
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system import sanity_check
from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.context.system.system_utils import SystemUtils


class ComponentTypeError(Exception):
    pass


class EmptySystemException(Exception):
    pass


class ComponentFactory:
    """
    class used to handle components describing a system
    """

    def __init__(self,
                 system_template: SystemTemplate):
        self.system_template = system_template
        self.shared_children_list: List[Component] = []
        self.uniq_id = 0

    def build(self):
        """
        Function generating a system from a dictionary describing a system
        :return: the single root node of tree representing the system.
        """
        # checking if tree is empty
        if not self.system_template.architecture_entry_list:
            message_exception = "The system architecture tree is empty."
            logging.exception("The system architecture tree is empty.")
            raise EmptySystemException(message_exception)

        # find root component, root component is the first line of the architecture tab.
        root_architecture_entry = \
            self.system_template.architecture_entry_list[0]

        component_list = self._add_component(root_architecture_entry,
                                             [],
                                             self.system_template.custom_children_logic_path)

        root_node = component_list[0]

        if not sanity_check.run(self.system_template, root_node):
            logging.warning('The input file and simulated tree do not match perfectly, please check the warnings.')

        return root_node

    def _add_component(self,
                       architecture_entry: ArchitectureEntry,
                       parents: List[Component],
                       custom_children_logic_path: str):
        """
        Recursive method to build the system graph.
        :param architecture_entry: The architecture_entry of the component to add in the system.
        :param parents: list of component parents of the component to add.
        :param custom_children_logic_path: string defining the path toward an optional python file containing custom
         children logic
        :return: List of the sibling components.
        """

        if architecture_entry.component_type_str == ComponentType.BASIC:
            return self._add_basics(architecture_entry,
                                    parents)

        elif architecture_entry.component_type_str == ComponentType.COMPOUND:
            return self._add_compound(architecture_entry,
                                      parents,
                                      custom_children_logic_path)
        else:
            message_exception = f"{architecture_entry.component_name} wrong type of component for component."
            logging.exception(message_exception)
            raise ComponentTypeError(message_exception)

    def _add_basics(self,
                    architecture_entry: ArchitectureEntry,
                    parents: List[Component]):
        """
        Part of the recursive logic from the _add_component.
        This methods creates a list of sibling Basic Components and return it.
        :param architecture_entry
        :param parents: the parents component for each of the basic components.
        :return: a list of Basics.
        """

        list_of_mrus__group = self.get_list_of_mru_from_list_of_strings(self.system_template.mru_list,
                                                                        architecture_entry.in_mru_str_list)

        list_of_mrus__trigger = self.get_list_of_mru_from_list_of_strings(self.system_template.mru_list,
                                                                          architecture_entry.trigger_mru_str_list)

        list_of_basic = []
        for component_number in range(0, int(architecture_entry.component_number)):
            self.uniq_id = self.uniq_id + 1
            for failure_mode_assignments in self.system_template.failure_mode_assignments_list:
                if failure_mode_assignments.component_name == architecture_entry.component_name:
                    list_of_basic.append(Basic(self.uniq_id,
                                               architecture_entry.component_name,
                                               component_number,
                                               parents.copy(),
                                               failure_mode_assignments.failure_mode,
                                               list_of_mrus__trigger,
                                               list_of_mrus__group)
                                         )
        return list_of_basic

    @staticmethod
    def get_list_of_mru_from_list_of_strings(mru_list,
                                             mru_list_of_names):
        list_of_mrus = [mru for mru in mru_list for mru_name in mru_list_of_names if mru.name == mru_name]
        return list_of_mrus

    def _add_compound(self,
                      architecture_entry: ArchitectureEntry,
                      parents: List[Component],
                      custom_children_logic_path: str):
        """
        Part of the recursive logic from the _add_component.
        This methods instantiates a new list of Compounds and their associated children.
        If the children are shared with multiple parents, they are created and stored in the shared_children_list.
        If the shared children has been created, then the compound is added to the list of parents.
        If the children is not shared (normal), then the children are created by calling the _add_component method.
        :param architecture_entry: the line describing the component within the input file
        :param parents: the list of parents.
        :param custom_children_logic_path: string defining the path toward an optional python file containing custom
        :return: list of the sibling compounds.
        """

        list_of_mru__trigger = self.get_list_of_mru_from_list_of_strings(self.system_template.mru_list,
                                                                         architecture_entry.trigger_mru_str_list)

        list_compounds = []
        for component_number in range(0, int(architecture_entry.component_number)):
            self.uniq_id = self.uniq_id + 1

            children_logic = ChildrenLogicFactory.build(architecture_entry.children_logic_str,
                                                        custom_children_logic_path)

            compound = Compound(self.uniq_id,
                                architecture_entry.component_name,
                                component_number,
                                parents,
                                children_logic,
                                list_of_mru__trigger)
            list_compounds.append(compound)

            for child_name in architecture_entry.children_name_list:
                if SystemUtils.is_string_containing_parenthesis(child_name):
                    shared_child_list = self.find_shared_children(compound,
                                                                  child_name,
                                                                  self.shared_children_list)
                    if shared_child_list:
                        for shared_child in shared_child_list:
                            shared_child.add_parent(compound)
                        component_list = shared_child_list
                    else:
                        architecture_entry_child = self.system_template.find_architecture_entry(
                            SystemUtils.extract_name_of_function_from_string(child_name))
                        if architecture_entry_child is None:
                            logging.warning(
                                f"child name {child_name} of the component {architecture_entry.component_name} not found")
                        component_list = self._add_component(architecture_entry_child,
                                                             [compound],
                                                             custom_children_logic_path)
                        self.shared_children_list.extend(component_list)
                else:
                    architecture_entry_child = self.system_template.find_architecture_entry(child_name)
                    if architecture_entry_child is None:
                        logging.warning(
                            f"child name {child_name} of the component {architecture_entry.component_name} not found")
                    component_list = self._add_component(architecture_entry_child, [compound], custom_children_logic_path)
                compound.add_children_list(component_list)
        return list_compounds

    @classmethod
    def find_shared_children(cls,
                             compound,
                             shared_child_name_str,
                             shared_children_list):
        """
        Find the shared child from the shared_children_list based on its name and its compound parent.
        :param compound: The parent compound of the child.
        :param shared_child_name_str: the name of the shared child (contains parenthesis)
        :param shared_children_list: the list of shared children already instantiated.
        :return: The shared child if is the list. None otherwise.
        """
        lowest_common_parent_name = SystemUtils.extract_arguments_within_parenthesis(shared_child_name_str)

        compound_lowest_common_parent = SystemUtils.find_first_lowest_level_ancestor_with_name(compound,
                                                                                               lowest_common_parent_name)

        shared_children = filter(
            lambda component: component.name == SystemUtils.extract_name_of_function_from_string(shared_child_name_str),
            shared_children_list)

        shared_children_found = []
        for shared_child in shared_children:
            shared_child_common_parent = SystemUtils.find_first_lowest_level_ancestor_with_name(shared_child,
                                                                                                lowest_common_parent_name)
            if compound_lowest_common_parent == shared_child_common_parent:
                shared_children_found.append(shared_child)
        return shared_children_found


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result, RCAResult


class test_E2E_RCA(unittest.TestCase):
    # the Root Cause Analysis (RCA) makes it possible to save the state of the system when a selected component enters 
    # a given status (both paramters specified in the input). Tests checking if the RCA feature detects and dumps the 
    # results correctly are defined here - they compare the "summary" output to the "rca" output. Different input files 
    # are used to identify cases where any other feature (like inspections, MRU or shared children) could prevent the 
    # RCA to work properly.

    # WARNING: results size in the RCA sheet depend strongly on the number of performed simulations, as snapshots are 
    # appended to the results list every time the conditions are met. Therefore, performing 10 simulations instead of
    # 1 increases the results size by a factor 10. The feature is designed primarily for rare events, but it is not
    # enforced in any way, i.e., it is ultimately user's decision of how many snapshots they want to record.

    expected_result_scenario_1 = [Result(10, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES",),
                                  Result(10, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(100, "ROOT_0_1", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_1.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_1 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT__AND__system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    expected_result_scenario_2 = [Result(10, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(10, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(100, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(100, "A_0_2", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_2.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_2 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT_A__AND__system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    expected_result_scenario_3 = [Result(0, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(10, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(0, "ROOT_0_1", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_3.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_3 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT__1OO2__system.xlsx",
         None,
         1,
         expected_result_scenario_3]

    expected_result_scenario_4 = [Result(0, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(10, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(0, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(100, "A_0_2", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_4.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_4 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT_A__1OO2__system.xlsx",
         None,
         1,
         expected_result_scenario_4]

    expected_result_scenario_5 = [Result(10, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(10, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(10, "S_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(100, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(100, "S_0_4", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "B_0_5", "C_0_3", "S_0_4"]:
        expected_result_scenario_5.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_5 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/shared__AND__AND__system.xlsx",
         None,
         1,
         expected_result_scenario_5]

    expected_result_scenario_6 = [Result(5, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(5, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(0, "B_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(0, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(0, "A_0_2", "*", "FAILED")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "B_0_3"]:
        expected_result_scenario_6.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_6 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT_A__AND__with_phases__v1__system.xlsx",
         None,
         1,
         expected_result_scenario_6]

    expected_result_scenario_7 = [Result(5, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(5, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(0, "B_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(50, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(50, "A_0_2", "*", "FAILED"),
                                  RCAResult(50, "ROOT_0_1", "P1", "FAILED"),
                                  RCAResult(50, "A_0_2", "P1", "FAILED"),
                                  RCAResult(0, "ROOT_0_1", "P2", "FAILED"),
                                  RCAResult(0, "A_0_2", "P2", "FAILED"),
                                  RCAResult(0, "ROOT_0_1", "SAFE", "FAILED"),
                                  RCAResult(0, "A_0_2", "SAFE", "FAILED")
                                  ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "B_0_3"]:
        expected_result_scenario_7.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_7 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT_A__AND__with_phases__v2__system.xlsx",
         None,
         1,
         expected_result_scenario_7]

    expected_result_scenario_8 = [Result(5, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(5, "A_0_2", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(0, "B_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  RCAResult(50, "ROOT_0_1", "*", "FAILED"),
                                  RCAResult(50, "A_0_2", "*", "FAILED"),
                                  RCAResult(50, "ROOT_0_1", "P1", "FAILED"),
                                  RCAResult(50, "A_0_2", "P1", "FAILED"),
                                  RCAResult(0, "ROOT_0_1", "P2", "FAILED"),
                                  RCAResult(0, "A_0_2", "P2", "FAILED"),
                                  RCAResult(0, "ROOT_0_1", "SAFE", "FAILED"),
                                  RCAResult(0, "A_0_2", "SAFE", "FAILED"),
                                  RCAResult(150, "ROOT_0_1", "*", "RUNNING"),
                                  RCAResult(100, "A_0_2", "*", "RUNNING"),
                                  RCAResult(100, "ROOT_0_1", "P1", "RUNNING"),
                                  RCAResult(100, "A_0_2", "P1", "RUNNING"),
                                  RCAResult(0, "ROOT_0_1", "P2", "RUNNING"),
                                  RCAResult(0, "A_0_2", "P2", "RUNNING"),
                                  RCAResult(50, "ROOT_0_1", "SAFE", "RUNNING"),
                                  RCAResult(0, "A_0_2", "SAFE", "RUNNING")
                                  ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "B_0_3"]:
        expected_result_scenario_8.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_8 = \
        ["./availsim4core/test/E2E/input/RCA/simulation.xlsx",
         "./availsim4core/test/E2E/input/RCA/ROOT_A__AND__with_phases__v3__system.xlsx",
         None,
         1,
         expected_result_scenario_8]

    param_scenario = [
        scenario_1,
        scenario_2,
        scenario_3,
        scenario_4,
        scenario_5,
        scenario_6,
        scenario_7,
        scenario_8,
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/RCA/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.repair_event.minimal_replaceable_unit_order_repair_event import \
    MinimalReplaceableUnitOrderRepairEvent


class test_CEventGenerator(unittest.TestCase):

    def test_get_initial_c_events_basic_mru(self):
        mru_1 = MinimalReplaceableUnit("test_get_initial_c_events_basic_mru",
                                       DeterministicLaw(0),
                                       "IMMEDIATE",
                                       Status.FAILED,
                                       "ROOT")

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic_1 = Basic(2, "BASIC1", 0, [root_component], None, [mru_1], [mru_1])

        root_component.add_children_list([_basic_1])
        context = Context(root_component, PhaseManager(set(),set()), set())
        result = CEventGenerator.get_c_events_for_mrus(context,
                                                       context.root_component)

        mru_event1 = MinimalReplaceableUnitOrderRepairEvent(CEventPriority.MRU_ORDER_REPAIR_EVENT, context, _basic_1, mru_1)
        expected_results = set([mru_event1])

        self.assertEqual(result, expected_results)

    def test_get_initial_c_events_compound_multiple_mru(self):
        mru_1 = MinimalReplaceableUnit("test_get_initial_c_events_basic_mru_1",
                                       DeterministicLaw(0),
                                       "IMMEDIATE",
                                       Status.FAILED,
                                       "ROOT")

        mru_2 = MinimalReplaceableUnit("test_get_initial_c_events_basic_mru_2",
                                       DeterministicLaw(0),
                                       "IMMEDIATE",
                                       Status.FAILED,
                                       "ROOT")

        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _compound_1 = Compound(2, "COMPOUND1", 0, [root_component], And(), [mru_1, mru_2])
        _basic_1 = Basic(3, "BASIC1", 0, [_compound_1], None, [], [mru_1])
        _basic_2 = Basic(4, "BASIC2", 0, [_compound_1], None, [], [mru_1])

        root_component.add_children_list([_compound_1])
        _compound_1.add_children_list([_basic_1, _basic_2])

        context = Context(root_component, PhaseManager(set(),set()), set())
        result = CEventGenerator.get_c_events_for_mrus(context,
                                                       context.root_component)
        mru_event1 = MinimalReplaceableUnitOrderRepairEvent(CEventPriority.MRU_ORDER_REPAIR_EVENT, context, _compound_1, mru_1)
        mru_event2 = MinimalReplaceableUnitOrderRepairEvent(CEventPriority.MRU_ORDER_REPAIR_EVENT, context, _compound_1, mru_2)
        expected_results = set([mru_event1, mru_event2])

        self.assertEqual(result, expected_results)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_repair(unittest.TestCase):
    """
    class testing the repair feature
    """

    # testing the failure and repair of a unique component
    expected_result_scenario_1 = [Result(10, "DUMB_0_1", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    scenario_1_no_sensitivity_analysis_single_component = \
        ["./availsim4core/test/E2E/input/repair/testRepair_simulation.xlsx",
         "./availsim4core/test/E2E/input/repair/testRepair_scenario_1_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    # testing the failure and repair of many components, results are similar to a unique component as failure and repair are in parallel
    expected_result_scenario_2 = [Result(10, "DUMB_0_1", "NONE", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    scenario_2_no_sensitivity_analysis_single_component = \
        ["./availsim4core/test/E2E/input/repair/testRepair_simulation.xlsx",
         "./availsim4core/test/E2E/input/repair/testRepair_scenario_2_system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    param_scenario = [
        scenario_1_no_sensitivity_analysis_single_component,
        scenario_2_no_sensitivity_analysis_single_component,
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_mean_result in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file, system_file, sensitivity_file, expected_number_file_result,
                                 expected_mean_result)

    def _runner_E2E(self, simulation_file, system_file, sensitivity_file, expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/repair/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.component_tree.status import Status


class PhaseJumpTrigger:

    def __init__(self,
                 component_name: str,
                 component_status: Status,
                 from_phase: str,
                 to_phase: str):
        self.component_name = component_name
        self.component_status = component_status
        self.from_phase = from_phase
        self.to_phase = to_phase

    def __str__(self):
        return f"Phase Jump trigger on {self.component_name}, "\
        f"status {self.component_status} from phase {self.from_phase} to phase {self.to_phase}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.component_name, self.component_status,
                     self.from_phase, self.to_phase))

    def __eq__(self, other):
        if not isinstance(other, PhaseJumpTrigger):
            return NotImplemented
        return self.component_name == other.component_name and \
               self.component_status == other.component_status and \
               self.from_phase == other.from_phase and \
               self.to_phase == other.to_phase


class PhaseJumpTriggerFactory:
    
    @staticmethod
    def build(triggering_component_name: str,
              triggering_component_status_list: List[str],
              triggering_from_phase_list: List[str],
              triggering_to_phase: str,
              phase_list: List[Phase]) -> List[PhaseJumpTrigger]:
        return [PhaseJumpTrigger(
            triggering_component_name,
            Status(status),
            next((phase for phase in phase_list if phase.name == from_phase), None),
            next((phase for phase in phase_list if phase.name == triggering_to_phase), None)
        )
            for status in triggering_component_status_list
            for from_phase in triggering_from_phase_list]



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class ResultRecordEntry:

    def identifier(self):
        """
        This method defines an Identifier of the record entry.
        This is useful, in case we want to regroup the entry based on this identifier. {cf SimulationResults}
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

import numpy as np
import qmcpy as qp

from availsim4core.src.simulation.des_random_generator.quasi_monte_carlo_generator import QuasiMonteCarloGenerator
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure import FailureType, Failure
from availsim4core.src.context.system.probability_law.exponential_law import ExponentialLaw

class test_QuasiMonteCarloGenerator(unittest.TestCase):

    def setUp(self):
        """
        setUp is a function called by the unit tester to initialize the class
        """
        simulation_duration = 1
        exponential_law_value = 1
        self.failure_mode_1 = FailureMode("test_run_1",
                                     ExponentialLaw([exponential_law_value]),
                                     ExponentialLaw([exponential_law_value]),
                                     Failure(FailureType.DETECTABLE),
                                     [],
                                     None,
                                     [],
                                     None,
                                     'NEVER',
                                     [])
        self.failure_mode_2 = FailureMode("test_run_2",
                                     ExponentialLaw([exponential_law_value]),
                                     ExponentialLaw([exponential_law_value]),
                                     Failure(FailureType.DETECTABLE),
                                     [],
                                     None,
                                     [],
                                     None,
                                     'NEVER',
                                     [])

        self.components_list = [Basic(1, "test_basic_1", 0, [], self.failure_mode_1, [], []),
                                Basic(2, "test_basic_2", 0, [], self.failure_mode_2, [], [])]
        self.qmc_generator = QuasiMonteCarloGenerator(self.components_list, simulation_duration, randomize=False, seed=0, graycode=True, multiplier=1)

    def test_generate_single_sample(self):
        expected_result_0 = qp.Sobol((6*1), randomize=False, seed=0, graycode=True).gen_samples(n_min=0, n_max=1)[0]
        expected_result_5 = qp.Sobol((6*1), randomize=False, seed=0, graycode=True).gen_samples(n_min=5, n_max=6)[0]
        expected_result_10 = qp.Sobol((6*1), randomize=False, seed=0, graycode=True).gen_samples(n_min=10, n_max=11)[0]

        np.testing.assert_almost_equal(self.qmc_generator.generate_single_sample(0), expected_result_0)
        np.testing.assert_almost_equal(self.qmc_generator.generate_single_sample(5), expected_result_5)
        np.testing.assert_almost_equal(self.qmc_generator.generate_single_sample(10),expected_result_10)

    def test_get_number_of_sample_failures(self):
        self.assertEqual(self.qmc_generator.get_number_of_sample_failures(), 3)

        # Empty components list
        result = QuasiMonteCarloGenerator([], 1).get_number_of_sample_failures()
        self.assertEqual(result, 0)

        # Components with and without failure modes
        qmc_gen = QuasiMonteCarloGenerator([Basic(1, "test_basic_1", 0, [], self.failure_mode_1, [], []),
                                            Basic(1, "test_basic_1", 0, [], self.failure_mode_1, [], [])], 3, multiplier=3)
        result = qmc_gen.get_number_of_sample_failures()
        self.assertEqual(result, 11)
        # Explanation: number of sample failures is the duration of simulation divided by the lowest mean time to fail defined in the system multiplied by
        # the multplier provided as the constructor argument. 
        # In this case, the simulation duration is 3, the lowest MTTF is equal to 1 and the multiplier is equal to 3, therefore the expected result is 9. 


    def test_set_uniform_samples_for_quasi_monte_carlo(self):
        sequence = [1, 2, 3, 4, 5, 6]
        self.qmc_generator.set_ttfs_of_failure_modes(sequence)

        self.assertEqual(self.components_list[0].failure_mode.uniform_samples_for_quasi_monte_carlo, [1,2,3])
        self.assertEqual(self.components_list[1].failure_mode.uniform_samples_for_quasi_monte_carlo, [4,5,6])


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List
import copy

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit


class Basic(Component):
    """
    Leaf element of the Component tree.
    It defines the 'basic' element of the system which is susceptible of failure.
    """
    __slots__ = 'failure_mode', 'list_of_mru_group'

    def __init__(self,
                 global_id: int,
                 name: str,
                 local_id: int,
                 parents: List[Compound],
                 failure_mode: FailureMode,
                 list_of_mru_trigger: List[MinimalReplaceableUnit],
                 list_of_mru_group: List[MinimalReplaceableUnit]
                 ):
        super().__init__(global_id, name, local_id, parents, list_of_mru_trigger)
        self.failure_mode = copy.deepcopy(failure_mode)
        self.list_of_mru_group = list_of_mru_group

    def __str__(self, level=0) -> str:
        return ("\t" * level +
                f"{self.name}_{self.local_id}"
                f" :uniq_id:{self.global_id}"
                f" :parents: {self.get_parents()}"
                f" :status: {self.status.name}"
                f" :failure mode: {self.failure_mode}"
                f" :mru: {self.list_of_mru_group} -- {self.list_of_mru_trigger}\n")

    def __repr__(self):
        return f"{self.name}_{self.local_id} uniq_id: {self.global_id}"

    def to_set(self):
        return set([self])

    def update_status(self,
                      new_status: Status,
                      timestamp: float,
                      phase: Phase,
                      description: str,
                      context: Context):
        return self._update_status(new_status, timestamp, phase, description, context)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.component_tree.status import Status


class ChildrenLogic:
    """
    Class which defines the Logic between a Component and its Children.
    """

    def __repr__(self):
        return f"Not defined"

    def evaluate(self, list_of_children: List[Component], context: Context) -> Status:
        """
        Given a list of Children component, this method evaluates the Status.
        :param list_of_children: The list of Children of the node to evaluate.
        :param context: the context of the simulation (structure containing almost everything)
        :return: The Status of the evaluated node.
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.simulation.simulation import Simulation


class QuickParse(Simulation):

    def __init__(self):
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.reader.xlsx.monte_carlo_reader import MonteCarloReader
from availsim4core.src.simulation.monte_carlo import MonteCarlo


class test_MonteCarloReader(unittest.TestCase):

    def test_build(self):
        line = {'SIMULATION_TYPE': 'MONTE_CARLO',
                'MIN_NUMBER_OF_SIMULATION': 100,
                'MAX_NUMBER_OF_SIMULATION': 100,
                'CONVERGENCE_MARGIN': 2,
                'MAX_EXECUTION_TIME': 600,
                'SEED': 1,
                'DIAGNOSTICS': 'summary',
                'SIMULATION_DURATION': 1000}

        expected_result = MonteCarlo(100, 100, 2, 600, 1, ["SUMMARY"], 1000)
        result = MonteCarloReader.build(line)
        self.assertEqual(expected_result, result)

    def test_build_with_list_of_diagnosis(self):
        line = {'SIMULATION_TYPE': 'MONTE_CARLO',
                'MIN_NUMBER_OF_SIMULATION': 100,
                'MAX_NUMBER_OF_SIMULATION': 100,
                'CONVERGENCE_MARGIN': 2,
                'MAX_EXECUTION_TIME': 600,
                'SEED': 1,
                'DIAGNOSTICS': "summary,last_timeline",
                'SIMULATION_DURATION': 1000}

        expected_result = MonteCarlo(100, 100, 2, 600, 1, ["SUMMARY","LAST_TIMELINE"], 1000)
        result = MonteCarloReader.build(line)
        self.assertEqual(expected_result, result)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_probabilityFailure(unittest.TestCase):
    """
    class used to recover analytical results: the probability of failure of simple systems after a given time,
    following an exponential distribution law of failure.
    """

    expected_result_scenario_1 = [
        Result(0.631, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", tolerance=6)
    ]
    scenario_1_without_sensitivity_analysis_1000simulations = \
        ["./availsim4core/test/E2E/input/convergence/N1000_simulation.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    # the theoretical values for the following test are (1 - exp(-1/MTTF))**N
    #  0 :  0.6321205588285577
    #  1 :  0.3934693402873666
    #  2 :  0.8646647167633873
    #  3 :  0.8646647167633873
    #  4 :  0.6321205588285577
    #  5 :  0.9816843611112658
    #  6 :  0.950212931632136
    #  7 :  0.7768698398515702
    #  8 :  0.9975212478233336
    #  9 :  0.9816843611112658
    #  10 :  0.8646647167633873
    #  11 :  0.9996645373720975
    #  12 :  0.9932620530009145
    #  13 :  0.9179150013761012
    #  14 :  0.9999546000702375

    # results are hardcoded based on a 1000 simulation run.
    expected_result_scenario_2 = [
        Result(0.631, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=0, tolerance=6),
        Result(0.407, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=1, tolerance=6),
        Result(0.872, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=2, tolerance=6),
        Result(0.871, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=3, tolerance=6),
        Result(0.645, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=4, tolerance=6),
        Result(0.984, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=5, tolerance=6),
        Result(0.953, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=6, tolerance=6),
        Result(0.791, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=7, tolerance=6),
        Result(0.999, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=8, tolerance=6),
        Result(0.982, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=9, tolerance=6),
        Result(0.868, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=10, tolerance=6),
        Result(1, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=11, tolerance=6),
        Result(0.989, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=12, tolerance=6),
        Result(0.925, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=13, tolerance=6),
        Result(1, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=14, tolerance=6),
    ]
    scenario_2_with_sensitivity_analysis_1000simulations = \
        ["./availsim4core/test/E2E/input/convergence/N1000_simulation.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_system.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_sensitivityAnalysis.xlsx",
         15, expected_result_scenario_2]

    # results are hardcoded based on a 10000 simulations run.

    expected_result_scenario_3 = [
        Result(0.6261, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=0, tolerance=6),
        Result(0.3896, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=1, tolerance=6),
        Result(0.8650, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=2, tolerance=6),
        Result(0.8629, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=3, tolerance=6),
        Result(0.6284, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=4, tolerance=6),
        Result(0.9818, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=5, tolerance=6),
        Result(0.9489, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=6, tolerance=6),
        Result(0.7771, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=7, tolerance=6),
        Result(0.9967, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=8, tolerance=6),
        Result(0.9816, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=9, tolerance=6),
        Result(0.8658, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=10, tolerance=6),
        Result(0.9995, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=11, tolerance=6),
        Result(0.9930, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=12, tolerance=6),
        Result(0.9158, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=13, tolerance=6),
        Result(0.9999, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=14, tolerance=6),
    ]
    scenario_3_with_sensitivity_analysis_10000simulations = \
        ["./availsim4core/test/E2E/input/convergence/N10000_simulation.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_system.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_sensitivityAnalysis.xlsx",
         15, expected_result_scenario_3]

    # results are hardcoded based on a 100000 simulation run.
    expected_result_scenario_4 = [
        Result(0.63013, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=0, tolerance=6),
        Result(0.39196, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=1, tolerance=6),
        Result(0.86527, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=2, tolerance=6),
        Result(0.86316, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=3, tolerance=6),
        Result(0.63064, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=4, tolerance=6),
        Result(0.98191, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=5, tolerance=6),
        Result(0.95039, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=6, tolerance=6),
        Result(0.77644, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=7, tolerance=6),
        Result(0.99728, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=8, tolerance=6),
        Result(0.98176, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=9, tolerance=6),
        Result(0.86445, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=10, tolerance=6),
        Result(0.99963, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=11, tolerance=6),
        Result(0.99335, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=12, tolerance=6),
        Result(0.91651, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=13, tolerance=6),
        Result(0.99995, "DUMB_0_1", "NONE", "BLIND_FAILED", "_MEAN_OCCURRENCES", file_number=14, tolerance=6),
    ]
    scenario_4_with_sensitivity_analysis_100000simulations = \
        ["./availsim4core/test/E2E/input/convergence/N100000_simulation.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_system.xlsx",
         "./availsim4core/test/E2E/input/convergence/convergence_test_sensitivityAnalysis.xlsx",
         15, expected_result_scenario_4]

    param_scenario = [
        scenario_1_without_sensitivity_analysis_1000simulations,
        scenario_2_with_sensitivity_analysis_1000simulations,
        # scenario_3_with_sensitivity_analysis_10000simulations,
        # scenario_4_with_sensitivity_analysis_100000simulations
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        # random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_mean_result in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file, system_file, sensitivity_file, expected_number_file_result,
                                 expected_mean_result)

    def _runner_E2E(self, simulation_file, system_file, sensitivity_file, expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/convergence/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)
            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set

from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.inspection_event.reevaluate_order_all_inspection_event import \
    ReevaluateOrderAllInspectionEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class EndInspectionEvent(BasicBEvent):
    """
    Defines the End of an inspection event.
    Attributes:
        - absolute_occurrence_time: date when the inspection event will be started.
        - component: associate component to inspect.
        - event: root event generating the EndInspectionEvent
        - failure_mode: failure mode which triggered this inspection event.
    """
    __slots__ = 'failure_mode', 'event'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 event: Event,
                 failure_mode: FailureMode):
        super().__init__(absolute_occurrence_time, context, basic,
                         priority=BEventPriority.END_INSPECTION_EVENT)
        self.event = event
        self.failure_mode = failure_mode

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.basic, self.failure_mode,
                     self.event))

    def __eq__(self, other):
        if not isinstance(other, EndInspectionEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.failure_mode == other.failure_mode and \
               self.event == other.event

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"at t ::{self.absolute_occurrence_time}, "
                f"FailureMode ::{self.failure_mode}\n")

    def execute(self):
        return self.basic.update_status(Status.RUNNING,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,  # or phase of the failure mode?
                                        "EndInspectionEvent",
                                        self.context)

    def generate_c_event(self, **kwargs) -> Set[CEvent]:
        """
        function which parses the list of C events and updates it:
        when an inspection is performed, the C events related to a given (component - failure mode)
        have to be removed, only the next inspection has to be ordered again
        In StartInspectionEvent we remove those events.
        In EndInspectionEvent we generate new ones.
        """

        set_of_c_events = set()

        set_of_c_events.update(
            CEventGenerator.get_c_events_for_failures(
                self.absolute_occurrence_time,
                self.context,
                self.basic
            )
        )

        set_of_c_events.add(
            ReevaluateOrderAllInspectionEvent(
                priority=CEventPriority.REEVALUATE_ORDER_ALL_INSPECTION_EVENTS,
                context=self.context,
                component=self.basic,
                event=self.event
            )
        )

        return set_of_c_events



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
import pandas as pd

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.results.results import Results
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.statistics.critical_failure_paths import CriticalFailurePaths


class XLSXResultsExporterCriticalFailurePaths(ResultExporter):

    def __init__(self, root_component_for_the_analysis: str):
        self.root_component_for_the_analysis = root_component_for_the_analysis

    def export(self,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str):
        f"""
        Exports results.
        :param context {Context} of the result to export.
        :param output_folder folder to export result.
        :param analysis {Analysis} the analysis which have been proceed.
        :param results {Results} the simulation results.
        :param execution_time time of the simulation.
        :param unique_output_file_identifier_str to uniquely identify group of output files. Currently the date/time at exporting is used as a label.
        """
        output_file = ResultExporter.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        logging.info(f"Exporting the critical failure path from the node {self.root_component_for_the_analysis}")
        cfp = CriticalFailurePaths(context, self.root_component_for_the_analysis)
        critical_failure_paths_names = []
        for path in cfp.critical_failure_paths:
            path_names = []
            for component in path:
                #TODO add some information about the failure mode ? (blind, dectable etc ?) or forcing an additional
                # export with this information, like the list of component export ?
                path_names.append(
                    f"{component.name}_{component.local_id}_{component.global_id}"
                )
            critical_failure_paths_names.append(path_names)

        cfp_df = pd.DataFrame(critical_failure_paths_names)
        cfp_df.to_excel(writer, sheet_name='RESULTS_CRITICAL_FAILURE_PATHS', index=False)
        writer.save()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.timeline.timeline import Timeline


class Results:
    """
    The interface for the simulation results.
    """

    def update_with_des_results(self,
                                execution_metrics: dict,
                                timeline: Timeline):
        """
        Update the results given the metric about the code execution and the full timeline of the discrete event simulation.
        :return:
        """
        pass

    def evaluate_result(self):
        """
        Evaluates the results at the end of all the simulation process, when all the des have been proceed.
        :return:
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import Dict

from availsim4core.src.reader.xlsx.monte_carlo_reader import MonteCarloReader
from availsim4core.src.reader.xlsx.monte_carlo_reader import SimulationMonteCarloColumn
from availsim4core.src.simulation.quasi_monte_carlo import QuasiMonteCarlo


class QuasiMonteCarloReader(MonteCarloReader):
    """
    Specific reader for the Quasi-Monte Carlo simulation.
    """

    @classmethod
    def build(cls, simulation_parameters: Dict):
        return QuasiMonteCarlo(
            int(simulation_parameters[SimulationMonteCarloColumn.MINIMUM_NUMBER_OF_SIMULATION]),
            int(simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_NUMBER_OF_SIMULATION]),
            float(simulation_parameters[SimulationMonteCarloColumn.CONVERGENCE_MARGIN]),
            float(simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_EXECUTION_TIME]),
            int(simulation_parameters[SimulationMonteCarloColumn.SEED]),
            cls.extract_from_str_list(simulation_parameters[SimulationMonteCarloColumn.DIAGNOSTICS]),
            float(simulation_parameters[SimulationMonteCarloColumn.DURATION])
        )


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from datetime import datetime
from functools import partial
from multiprocessing import Pool

from availsim4core import configuration
from availsim4core.src.analysis import Analysis
from availsim4core.src.reader.xlsx.simulation_reader import SimulationReader
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateReader
from availsim4core.src.runner.htcondor_runner import HTCondorRunner
from availsim4core.src.runner.local_runner import LocalRunner
from availsim4core.src.sensitivity_analysis.sensitivity_analysis import SensitivityAnalysis


def start(path_simulation: str,
          path_system: str,
          path_sensitivity_analysis: str,
          output_folder: str,
          HTCondor: bool = False,
          HTCondor_extra_argument: str = "",
          nb_processes: int = 1,
          custom_children_logic_path: str = None):
    """
    Starts the availsim simulation.
    If `HTCondor` is true it starts the software on the HTCondor cluster.
    """

    initial_simulation = SimulationReader.generate_simulation(path_simulation)

    initial_system_template = SystemTemplateReader.generate_system_template(path_system,
                                                                            custom_children_logic_path)

    analysis_list = [Analysis(0,
                              initial_system_template,
                              initial_simulation)]

    if path_sensitivity_analysis:
        # init the list of systems to be simulated

        sensitivity_analysis = SensitivityAnalysis(initial_simulation, initial_system_template)
        analysis_list = sensitivity_analysis.generate_analysis_list(path_sensitivity_analysis)

    if HTCondor:
        HTCondorRunner.run(output_folder,
                           analysis_list,
                           HTCondor_extra_argument)
    else:
        if nb_processes == 1:
            for analysis in analysis_list:
                LocalRunner.run(output_folder,
                                analysis)
        else:
            partial_function = partial(LocalRunner.run,
                                       output_folder)
            with Pool(processes=nb_processes) as pool:
                pool.map(partial_function, analysis_list)


def main():
    """
    Entry point of Availsim.
    """

    (simulation,
     system,
     sensitivity_analysis,
     output_folder,
     HTCondor,
     HTCondor_extra_argument,
     nb_processes,
     custom_children_logic_path) = configuration.init()

    logging.info(f"- Start simulating -")

    start_time = datetime.now()

    start(simulation,
          system,
          sensitivity_analysis,
          output_folder,
          HTCondor,
          HTCondor_extra_argument,
          nb_processes,
          custom_children_logic_path)

    execution_time = (datetime.now() - start_time).total_seconds()
    logging.info(f"Total Availsim execution time: {execution_time:.2f} s")



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.failure_event.order_failure_event import \
    OrderFailureEvent
from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_jump_phase_event import \
    OrderJumpPhaseEvent

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.component import Component

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.c_event.repair_event.minimal_replaceable_unit_order_repair_event import \
    MinimalReplaceableUnitOrderRepairEvent


class CEventGenerator:

    @classmethod
    def generate_first_events(cls,
                              absolute_simulation_time: float,
                              context: Context):
        """
        Function generating the first C events when a simulation is started
        """

        set_of_c_events = set()

        set_of_c_events.update(
            cls.get_c_events_for_mrus(
                context,
                context.root_component
            )
        )
        set_of_c_events.update(
            cls.get_c_events_for_failures(
                absolute_simulation_time,
                context,
                context.root_component
            )
        )
        set_of_c_events.update(
            cls._generate_first__phase_jump_events(
                context
            )
        )
        return set_of_c_events

    @classmethod
    def get_c_events_for_mrus(cls,
                             context: Context,
                             component: Component):
        """
        Function generating mru C events of a component recursively,
        calling itself on children of the current component if any
        """
        set_of_c_events = set()
        for mru in component.list_of_mru_trigger:
            set_of_c_events.add(
                MinimalReplaceableUnitOrderRepairEvent(
                    priority=CEventPriority.MRU_ORDER_REPAIR_EVENT,
                    context=context,
                    component=component,
                    mru=mru
                )
            )
        for child_component in component.get_children():
            events = cls.get_c_events_for_mrus(context,
                                               child_component)
            set_of_c_events.update(events)
        return set_of_c_events

    @classmethod
    def get_c_events_for_failures(cls,
                                  absolute_simulation_time: float,
                                  context: Context,
                                  component: Component):
        """
        Function generating first failure B events of a component recursively,
        calling itself on children of the current component if any
        """
        set_of_events = set()
        for child_component in component.get_children():
            events = cls.get_c_events_for_failures(absolute_simulation_time,
                                                   context,
                                                   child_component)
            set_of_events.update(events)
        if not component.get_children():
            set_of_events.add(
                OrderFailureEvent(
                    priority=CEventPriority.ORDER_FAILURE_EVENT,
                    context=context,
                    component=component,
                    event=None,
                    failure_mode=component.failure_mode
                )
            )
        return set_of_events

    @classmethod
    def _generate_first__phase_jump_events(cls,
                                           context: Context):
        """
        Function generating first C events of a component recursively,
        calling itself on children of the current component if any
        """
        set_of_c_events = set()
        for phase_jump_trigger in context.phase_manager.phase_jump_trigger_set:
            set_of_c_events.add(
                OrderJumpPhaseEvent(
                    priority=CEventPriority.ORDER_JUMP_PHASE_EVENT,
                    context=context,
                    phase_jump_trigger=phase_jump_trigger
                )
            )

        return set_of_c_events